import io

from lxml import etree

from tests.utils import DummyTransport, assert_nodes_equal, load_xml
from zeep import xsd


def get_transport():
    transport = DummyTransport()
    transport.bind(
        'http://schemas.xmlsoap.org/soap/encoding/',
        load_xml(io.open('tests/wsdl_files/soap-enc.xsd', 'r').read().encode('utf-8')))
    return transport


def test_simple_type():
    schema = xsd.Schema(load_xml("""
    <xsd:schema
        xmlns:xsd="http://www.w3.org/2001/XMLSchema"
        xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/"
        xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
        targetNamespace="http://tests.python-zeep.org/tns">
      <xsd:import namespace="http://schemas.xmlsoap.org/soap/encoding/"/>
      <xsd:complexType name="ArrayOfString">
        <xsd:complexContent>
          <xsd:restriction base="SOAP-ENC:Array">
            <xsd:attribute ref="SOAP-ENC:arrayType" wsdl:arrayType="xsd:string[]"/>
          </xsd:restriction>
        </xsd:complexContent>
      </xsd:complexType>
    </xsd:schema>
    """), transport=get_transport())

    ArrayOfString = schema.get_type('ns0:ArrayOfString')
    print(ArrayOfString.__dict__)

    value = ArrayOfString(['item', 'and', 'even', 'more', 'items'])

    node = etree.Element('document')
    ArrayOfString.render(node, value)

    expected = """
        <document>
            <item xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="xs:string">item</item>
            <item xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="xs:string">and</item>
            <item xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="xs:string">even</item>
            <item xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="xs:string">more</item>
            <item xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:type="xs:string">items</item>
        </document>
    """  # noqa

    assert_nodes_equal(expected, node)


def test_complex_type():
    schema = xsd.Schema(load_xml("""
    <xsd:schema
        xmlns:xsd="http://www.w3.org/2001/XMLSchema"
        xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/"
        xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
        xmlns:tns="http://tests.python-zeep.org/tns"
        targetNamespace="http://tests.python-zeep.org/tns">
      <xsd:import namespace="http://schemas.xmlsoap.org/soap/encoding/"/>

      <xsd:complexType name="ArrayObject">
        <xsd:sequence>
          <xsd:element name="attr_1" type="xsd:string"/>
          <xsd:element name="attr_2" type="xsd:string"/>
        </xsd:sequence>
      </xsd:complexType>
      <xsd:complexType name="ArrayOfObject">
        <xsd:complexContent>
          <xsd:restriction base="SOAP-ENC:Array">
            <xsd:sequence>
              <xsd:element name="obj" type="tns:ArrayObject" minOccurs="0" maxOccurs="unbounded"/>
            </xsd:sequence>
            <xsd:attribute ref="SOAP-ENC:arrayType" wsdl:arrayType="tns:ArrayObject[]"/>
          </xsd:restriction>
        </xsd:complexContent>
      </xsd:complexType>
    </xsd:schema>
    """), transport=get_transport())

    ArrayOfObject = schema.get_type('ns0:ArrayOfObject')
    ArrayObject = schema.get_type('ns0:ArrayObject')

    value = ArrayOfObject([
        ArrayObject(attr_1='attr-1', attr_2='attr-2'),
        ArrayObject(attr_1='attr-3', attr_2='attr-4'),
        ArrayObject(attr_1='attr-5', attr_2='attr-6'),
    ])

    node = etree.Element('document')
    ArrayOfObject.render(node, value)

    expected = """
        <document>
            <obj>
                <attr_1>attr-1</attr_1>
                <attr_2>attr-2</attr_2>
            </obj>
            <obj>
                <attr_1>attr-3</attr_1>
                <attr_2>attr-4</attr_2>
            </obj>
            <obj>
                <attr_1>attr-5</attr_1>
                <attr_2>attr-6</attr_2>
            </obj>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_complex_type_without_name():
    schema = xsd.Schema(load_xml("""
    <xsd:schema
        xmlns:xsd="http://www.w3.org/2001/XMLSchema"
        xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/"
        xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
        xmlns:tns="http://tests.python-zeep.org/tns"
        targetNamespace="http://tests.python-zeep.org/tns">
      <xsd:import namespace="http://schemas.xmlsoap.org/soap/encoding/"/>

      <xsd:complexType name="ArrayObject">
        <xsd:sequence>
          <xsd:element name="attr_1" type="xsd:string"/>
          <xsd:element name="attr_2" type="xsd:string"/>
        </xsd:sequence>
      </xsd:complexType>
      <xsd:complexType name="ArrayOfObject">
        <xsd:complexContent>
          <xsd:restriction base="SOAP-ENC:Array">
            <xsd:attribute ref="SOAP-ENC:arrayType" wsdl:arrayType="tns:ArrayObject[]"/>
          </xsd:restriction>
        </xsd:complexContent>
      </xsd:complexType>
    </xsd:schema>
    """), transport=get_transport())

    ArrayOfObject = schema.get_type('ns0:ArrayOfObject')
    ArrayObject = schema.get_type('ns0:ArrayObject')

    value = ArrayOfObject([
        ArrayObject(attr_1='attr-1', attr_2='attr-2'),
        ArrayObject(attr_1='attr-3', attr_2='attr-4'),
        ArrayObject(attr_1='attr-5', attr_2='attr-6'),
    ])

    node = etree.Element('document')
    ArrayOfObject.render(node, value)

    expected = """
        <document>
            <ArrayObject>
                <attr_1>attr-1</attr_1>
                <attr_2>attr-2</attr_2>
            </ArrayObject>
            <ArrayObject>
                <attr_1>attr-3</attr_1>
                <attr_2>attr-4</attr_2>
            </ArrayObject>
            <ArrayObject>
                <attr_1>attr-5</attr_1>
                <attr_2>attr-6</attr_2>
            </ArrayObject>
        </document>
    """
    assert_nodes_equal(expected, node)
    data = ArrayOfObject.parse_xmlelement(node, schema)

    assert len(data._value_1) == 3
    assert data._value_1[0]['attr_1'] == 'attr-1'
    assert data._value_1[0]['attr_2'] == 'attr-2'
    assert data._value_1[1]['attr_1'] == 'attr-3'
    assert data._value_1[1]['attr_2'] == 'attr-4'
    assert data._value_1[2]['attr_1'] == 'attr-5'
    assert data._value_1[2]['attr_2'] == 'attr-6'


def test_soap_array_parse_remote_ns():
    transport = DummyTransport()
    transport.bind(
        'http://schemas.xmlsoap.org/soap/encoding/',
        load_xml(io.open('tests/wsdl_files/soap-enc.xsd', 'r').read().encode('utf-8')))

    schema = xsd.Schema(load_xml("""
        <?xml version="1.0"?>
        <xsd:schema
          xmlns:xsd="http://www.w3.org/2001/XMLSchema"
          xmlns:tns="http://tests.python-zeep.org/"
          xmlns:soapenc="http://schemas.xmlsoap.org/soap/encoding/"
          xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/"
          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
          targetNamespace="http://tests.python-zeep.org/"
          elementFormDefault="qualified">
          <xsd:import namespace="http://schemas.xmlsoap.org/soap/encoding/"/>
          <xsd:simpleType name="CountryCodeType">
            <xsd:restriction base="xsd:string">
              <xsd:length value="2"/>
              <xsd:pattern value="[a-zA-Z]{2}"/>
            </xsd:restriction>
          </xsd:simpleType>
          <xsd:complexType name="CountryItemType">
            <xsd:sequence>
              <xsd:element name="code" type="tns:CountryCodeType"/>
              <xsd:element name="name" type="xsd:string"/>
            </xsd:sequence>
          </xsd:complexType>
          <xsd:complexType name="CountriesArrayType">
            <xsd:complexContent>
              <xsd:restriction base="soapenc:Array">
                <xsd:attribute ref="soapenc:arrayType" wsdl:arrayType="tns:CountryItemType[]"/>
              </xsd:restriction>
            </xsd:complexContent>
          </xsd:complexType>
          <xsd:element name="countries" type="tns:CountriesArrayType"/>
        </xsd:schema>
    """), transport)

    doc = load_xml("""
      <countries
            SOAP-ENC:arrayType="ns1:CountryItemType[1]"
            xsi:type="ns1:CountriesArrayType"
            xmlns:ns1="http://tests.python-zeep.org/"
            xmlns:xsd="http://www.w3.org/2001/XMLSchema"
            xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        <item xsi:type="ns1:CountryItemType">
          <code xsi:type="ns1:CountryCodeType">NL</code>
          <name xsi:type="xsd:string">The Netherlands</name>
        </item>
      </countries>
    """)

    elm = schema.get_element('ns0:countries')
    data = elm.parse(doc, schema)

    assert data._value_1[0].code == 'NL'
    assert data._value_1[0].name == 'The Netherlands'


def test_wsdl_array_type():
    transport = DummyTransport()
    transport.bind(
        'http://schemas.xmlsoap.org/soap/encoding/',
        load_xml(io.open('tests/wsdl_files/soap-enc.xsd', 'r').read().encode('utf-8')))

    schema = xsd.Schema(load_xml("""
        <xsd:schema xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                    xmlns:tns="http://tests.python-zeep.org/"
                    xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
                    xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/"
                    targetNamespace="http://tests.python-zeep.org/"
                    elementFormDefault="qualified">
          <xsd:import namespace="http://schemas.xmlsoap.org/soap/encoding/"/>
          <xsd:complexType name="array">
            <xsd:complexContent>
              <xsd:restriction base="SOAP-ENC:Array">
                <xsd:attribute ref="SOAP-ENC:arrayType" wsdl:arrayType="tns:base[]"/>
              </xsd:restriction>
            </xsd:complexContent>
          </xsd:complexType>
          <xsd:complexType name="base">
            <xsd:sequence>
              <xsd:element minOccurs="0" name="item_1" type="xsd:string"/>
              <xsd:element minOccurs="0" name="item_2" type="xsd:string"/>
            </xsd:sequence>
          </xsd:complexType>
          <xsd:element name="array" type="tns:array"/>
        </xsd:schema>
    """), transport)
    array_elm = schema.get_element('{http://tests.python-zeep.org/}array')

    item_type = schema.get_type('{http://tests.python-zeep.org/}base')
    item_1 = item_type(item_1='foo_1', item_2='bar_1')
    item_2 = item_type(item_1='foo_2', item_2='bar_2')

    # array = array_elm([
    #     xsd.AnyObject(item_type, item_1),
    #     xsd.AnyObject(item_type, item_2),
    # ])

    array = array_elm([item_1, item_2])
    node = etree.Element('document')
    assert array_elm.signature(schema=schema) == 'ns0:array(ns0:array)'

    array_type = schema.get_type('ns0:array')
    assert array_type.signature(schema=schema) == (
        'ns0:array(_value_1: base[], arrayType: xsd:string, ' +
        'offset: ns1:arrayCoordinate, id: xsd:ID, href: xsd:anyURI, _attr_1: {})')
    array_elm.render(node, array)
    expected = """
        <document>
            <ns0:array xmlns:ns0="http://tests.python-zeep.org/">
                <base>
                    <ns0:item_1>foo_1</ns0:item_1>
                    <ns0:item_2>bar_1</ns0:item_2>
                </base>
                <base>
                    <ns0:item_1>foo_2</ns0:item_1>
                    <ns0:item_2>bar_2</ns0:item_2>
                </base>
            </ns0:array>
        </document>
    """
    assert_nodes_equal(expected, node)


def test_soap_array_parse():
    transport = DummyTransport()
    transport.bind(
        'http://schemas.xmlsoap.org/soap/encoding/',
        load_xml(io.open('tests/wsdl_files/soap-enc.xsd', 'r').read().encode('utf-8')))

    schema = xsd.Schema(load_xml("""
    <?xml version="1.0"?>
    <schema xmlns="http://www.w3.org/2001/XMLSchema"
            xmlns:tns="http://tests.python-zeep.org/"
            xmlns:soapenc="http://schemas.xmlsoap.org/soap/encoding/"
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
            targetNamespace="http://tests.python-zeep.org/"
            elementFormDefault="qualified">
      <import namespace="http://schemas.xmlsoap.org/soap/encoding/"/>
      <complexType name="FlagDetailsStruct">
          <sequence>
              <element name="Name">
                  <simpleType>
                      <restriction base="string">
                          <maxLength value="512"/>
                      </restriction>
                  </simpleType>
              </element>
              <element name="Value" type="string"/>
          </sequence>
      </complexType>
      <complexType name="FlagDetailsList">
          <complexContent>
              <restriction base="soapenc:Array">
                  <sequence>
                      <element
                        name="FlagDetailsStruct" type="tns:FlagDetailsStruct"
                        minOccurs="0" maxOccurs="unbounded"/>
                  </sequence>
                  <attribute ref="soapenc:arrayType" use="required"/>
              </restriction>
          </complexContent>
      </complexType>
      <element name="FlagDetailsList" type="tns:FlagDetailsList"/>
    </schema>
    """), transport)

    doc = load_xml("""
         <FlagDetailsList xmlns="http://tests.python-zeep.org/">
            <FlagDetailsStruct>
               <Name>flag1</Name>
               <Value>value1</Value>
            </FlagDetailsStruct>
            <FlagDetailsStruct>
               <Name>flag2</Name>
               <Value>value2</Value>
            </FlagDetailsStruct>
         </FlagDetailsList>
    """)

    elm = schema.get_element('ns0:FlagDetailsList')
    data = elm.parse(doc, schema)
    assert data.FlagDetailsStruct[0].Name == 'flag1'
    assert data.FlagDetailsStruct[0].Value == 'value1'
    assert data.FlagDetailsStruct[1].Name == 'flag2'
    assert data.FlagDetailsStruct[1].Value == 'value2'



from zeep.xsd.utils import create_prefixed_name


class Type(object):

    def __init__(self, qname=None, is_global=False):
        self.qname = qname
        self.name = qname.localname if qname else None
        self._resolved = False
        self.is_global = is_global

    def get_prefixed_name(self, schema):
        return create_prefixed_name(self.qname, schema)

    def accept(self, value):
        raise NotImplementedError

    def validate(self, value, required=False):
        return

    def parse_kwargs(self, kwargs, name, available_kwargs):
        value = None
        name = name or self.name

        if name in available_kwargs:
            value = kwargs[name]
            available_kwargs.remove(name)
            return {name: value}
        return {}

    def parse_xmlelement(self, xmlelement, schema=None, allow_none=True,
                         context=None):
        raise NotImplementedError(
            '%s.parse_xmlelement() is not implemented' % self.__class__.__name__)

    def parsexml(self, xml, schema=None):
        raise NotImplementedError

    def render(self, parent, value, xsd_type=None, render_path=None):
        raise NotImplementedError(
            '%s.render() is not implemented' % self.__class__.__name__)

    def resolve(self):
        raise NotImplementedError(
            '%s.resolve() is not implemented' % self.__class__.__name__)

    def extend(self, child):
        raise NotImplementedError(
            '%s.extend() is not implemented' % self.__class__.__name__)

    def restrict(self, child):
        raise NotImplementedError(
            '%s.restrict() is not implemented' % self.__class__.__name__)

    @property
    def attributes(self):
        return []

    @classmethod
    def signature(cls, schema=None, standalone=True):
        return ''


class UnresolvedType(Type):
    def __init__(self, qname, schema):
        self.qname = qname
        assert self.qname.text != 'None'
        self.schema = schema

    def __repr__(self):
        return '<%s(qname=%r)>' % (self.__class__.__name__, self.qname.text)

    def render(self, parent, value, xsd_type=None, render_path=None):
        raise RuntimeError(
            "Unable to render unresolved type %s. This is probably a bug." % (
                self.qname))

    def resolve(self):
        retval = self.schema.get_type(self.qname)
        return retval.resolve()


class UnresolvedCustomType(Type):

    def __init__(self, qname, base_type, schema):
        assert qname is not None
        self.qname = qname
        self.name = str(qname.localname)
        self.schema = schema
        self.base_type = base_type

    def __repr__(self):
        return '<%s(qname=%r, base_type=%r)>' % (
            self.__class__.__name__, self.qname.text, self.base_type)

    def resolve(self):
        base = self.base_type
        base = base.resolve()

        cls_attributes = {
            '__module__': 'zeep.xsd.dynamic_types',
        }

        from zeep.xsd.types.collection import UnionType  # FIXME
        from zeep.xsd.types.simple import AnySimpleType  # FIXME

        if issubclass(base.__class__, UnionType):
            xsd_type = type(self.name, (base.__class__,), cls_attributes)
            return xsd_type(base.item_types)

        elif issubclass(base.__class__, AnySimpleType):
            xsd_type = type(self.name, (base.__class__,), cls_attributes)
            return xsd_type(self.qname)

        else:
            xsd_type = type(self.name, (base.base_class,), cls_attributes)
            return xsd_type(self.qname)



from __future__ import print_function
import zeep
from zeep.transports import Transport

# Example using basic authentication with a webservice

transport_with_basic_auth = Transport(http_auth=('username', 'password'))

client = zeep.Client(
    wsdl='http://nonexistent?WSDL',
    transport=transport_with_basic_auth
)

client.wsdl.dump()



import copy
import logging

from lxml import etree

from zeep import exceptions
from zeep.exceptions import UnexpectedElementError
from zeep.utils import qname_attr
from zeep.xsd.const import NotSet, xsi_ns
from zeep.xsd.context import XmlParserContext
from zeep.xsd.elements.base import Base
from zeep.xsd.utils import max_occurs_iter, create_prefixed_name

logger = logging.getLogger(__name__)

__all__ = ['Element']


class Element(Base):
    def __init__(self, name, type_=None, min_occurs=1, max_occurs=1,
                 nillable=False, default=None, is_global=False, attr_name=None):

        if name is None:
            raise ValueError("name cannot be None", self.__class__)
        if not isinstance(name, etree.QName):
            name = etree.QName(name)

        self.name = name.localname if name else None
        self.qname = name
        self.type = type_
        self.min_occurs = min_occurs
        self.max_occurs = max_occurs
        self.nillable = nillable
        self.is_global = is_global
        self.default = default
        self.attr_name = attr_name or self.name
        # assert type_

    def __str__(self):
        if self.type:
            if self.type.is_global:
                return '%s(%s)' % (self.name, self.type.qname)
            else:
                return '%s(%s)' % (self.name, self.type.signature())
        return '%s()' % self.name

    def __call__(self, *args, **kwargs):
        instance = self.type(*args, **kwargs)
        if hasattr(instance, '_xsd_type'):
            instance._xsd_elm = self
        return instance

    def __repr__(self):
        return '<%s(name=%r, type=%r)>' % (
            self.__class__.__name__, self.name, self.type)

    def __eq__(self, other):
        return (
            other is not None and
            self.__class__ == other.__class__ and
            self.__dict__ == other.__dict__)

    def get_prefixed_name(self, schema):
        return create_prefixed_name(self.qname, schema)

    @property
    def default_value(self):
        value = [] if self.accepts_multiple else self.default
        return value

    def clone(self, name=None, min_occurs=1, max_occurs=1):
        new = copy.copy(self)

        if name:
            if not isinstance(name, etree.QName):
                name = etree.QName(name)
            new.name = name.localname
            new.qname = name
            new.attr_name = new.name

        new.min_occurs = min_occurs
        new.max_occurs = max_occurs
        return new

    def parse(self, xmlelement, schema, allow_none=False, context=None):
        """Process the given xmlelement. If it has an xsi:type attribute then
        use that for further processing. This should only be done for subtypes
        of the defined type but for now we just accept everything.

        """
        context = context or XmlParserContext()
        instance_type = qname_attr(xmlelement, xsi_ns('type'))
        xsd_type = None
        if instance_type:
            xsd_type = schema.get_type(instance_type, fail_silently=True)
        xsd_type = xsd_type or self.type
        return xsd_type.parse_xmlelement(
            xmlelement, schema, allow_none=allow_none, context=context)

    def parse_kwargs(self, kwargs, name, available_kwargs):
        return self.type.parse_kwargs(
            kwargs, name or self.attr_name, available_kwargs)

    def parse_xmlelements(self, xmlelements, schema, name=None, context=None):
        """Consume matching xmlelements and call parse() on each of them

        :param xmlelements: Dequeue of XML element objects
        :type xmlelements: collections.deque of lxml.etree._Element
        :param schema: The parent XML schema
        :type schema: zeep.xsd.Schema
        :param name: The name of the parent element
        :type name: str
        :param context: Optional parsing context (for inline schemas)
        :type context: zeep.xsd.context.XmlParserContext
        :return: dict or None

        """
        result = []
        num_matches = 0
        for _unused in max_occurs_iter(self.max_occurs):
            if not xmlelements:
                break

            # Workaround for SOAP servers which incorrectly use unqualified
            # or qualified elements in the responses (#170, #176). To make the
            # best of it we compare the full uri's if both elements have a
            # namespace. If only one has a namespace then only compare the
            # localname.

            # If both elements have a namespace and they don't match then skip
            element_tag = etree.QName(xmlelements[0].tag)
            if (
                element_tag.namespace and self.qname.namespace and
                element_tag.namespace != self.qname.namespace and
                schema.strict
            ):
                break

            # Only compare the localname
            if element_tag.localname == self.qname.localname:
                xmlelement = xmlelements.popleft()
                num_matches += 1
                item = self.parse(
                    xmlelement, schema, allow_none=True, context=context)
                result.append(item)
            else:
                # If the element passed doesn't match and the current one is
                # not optional then throw an error
                if num_matches == 0 and not self.is_optional:
                    raise UnexpectedElementError(
                        "Unexpected element %r, expected %r" % (
                            element_tag.text, self.qname.text))
                break

        if not self.accepts_multiple:
            result = result[0] if result else None
        return result

    def render(self, parent, value, render_path=None):
        """Render the value(s) on the parent lxml.Element.

        This actually just calls _render_value_item for each value.

        """
        if not render_path:
            render_path = [self.qname.localname]

        assert parent is not None
        self.validate(value, render_path)

        if self.accepts_multiple and isinstance(value, list):
            for val in value:
                self._render_value_item(parent, val, render_path)
        else:
            self._render_value_item(parent, value, render_path)

    def _render_value_item(self, parent, value, render_path):
        """Render the value on the parent lxml.Element"""
        if value is None or value is NotSet:
            if self.is_optional:
                return

            elm = etree.SubElement(parent, self.qname)
            if self.nillable:
                elm.set(xsi_ns('nil'), 'true')
            return

        node = etree.SubElement(parent, self.qname)
        xsd_type = getattr(value, '_xsd_type', self.type)

        if xsd_type != self.type:
            return value._xsd_type.render(node, value, xsd_type, render_path)
        return self.type.render(node, value, None, render_path)

    def validate(self, value, render_path=None):
        """Validate that the value is valid"""
        if self.accepts_multiple and isinstance(value, list):

            # Validate bounds
            if len(value) < self.min_occurs:
                raise exceptions.ValidationError(
                    "Expected at least %d items (minOccurs check)" % self.min_occurs,
                    path=render_path)
            elif self.max_occurs != 'unbounded' and len(value) > self.max_occurs:
                raise exceptions.ValidationError(
                    "Expected at most %d items (maxOccurs check)" % self.min_occurs,
                    path=render_path)

            for val in value:
                self._validate_item(val, render_path)
        else:
            if not self.is_optional and not self.nillable and value in (None, NotSet):
                raise exceptions.ValidationError(
                    "Missing element %s" % (self.name), path=render_path)

            self._validate_item(value, render_path)

    def _validate_item(self, value, render_path):
        if self.nillable and value in (None, NotSet):
            return

        try:
            self.type.validate(value, required=True)
        except exceptions.ValidationError as exc:
            raise exceptions.ValidationError(
                "The element %s is not valid: %s" % (self.qname, exc.message),
                path=render_path)

    def resolve_type(self):
        self.type = self.type.resolve()

    def resolve(self):
        self.resolve_type()
        return self

    def signature(self, schema=None, standalone=True):
        from zeep.xsd import ComplexType
        if self.type.is_global:
            value = self.type.get_prefixed_name(schema)
        else:
            value = self.type.signature(schema, standalone=False)

            if not standalone and isinstance(self.type, ComplexType):
                value = '{%s}' % value

        if standalone:
            value = '%s(%s)' % (self.get_prefixed_name(schema), value)

        if self.accepts_multiple:
            return '%s[]' % value
        return value



# -*- coding: utf-8 -*-
"""
==================================
Utils :mod:`transport_with_attach`
==================================
Utils to add basic MTOMS attachment

author: ellethee <luca800@gmail.com>
edit: remaudcorentin <remaudcorentin.dev@gmail.com>

based on http://stackoverflow.com/questions/35558812/how-to-send-multipart-related-requests-in-python-to-soap-server
then edited to fit specific requirements, correct mimetype and some more data into atachments headers
"""

from os.path import basename
import random
import string
import mimetypes
from email.mime.application import MIMEApplication
from email.encoders import encode_7or8bit
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from base64 import b64decode, b64encode
from lxml import etree
from zeep.transports import Transport
from zeep.wsdl.utils import etree_to_string
from zeep.xsd import builtins
from zeep.wsdl.bindings import http
from zeep.wsdl import wsdl

import codecs

import zeep.ns

from zeep import client
BOUND = "MTOM".center(40, "=")
# XOP_LINK = "http://www.w3.org/2004/08/xop/include"
FILETAG = 'xop:Include:'
ID_LEN = 16
func_getid = lambda N: ''.join(
    random.SystemRandom().choice(string.ascii_uppercase + string.digits)
    for _ in range(N))
mimetypes.init()

# I need it my WSDL uses it for the data part.
def xmlvalue(self, value):
    """Patch for xmlvalue"""
    if value.startswith(FILETAG):
        return value
    return b64encode(value)

def pythonvalue(self, value):
    """Patch for pythonvalue"""
    if value.startswith(FILETAG):
        return value
    return b64decode(value)

def patch_things():
    """Patches something"""
    # Let's patch the Base64Binary data type.
    builtins.Base64Binary.accepted_types += (etree.Element, )
    builtins.Base64Binary.xmlvalue = xmlvalue
    builtins.Base64Binary.pythonvalue = pythonvalue
    # Base64Binary patched.
    # Update NSMAP
    zeep.ns.XOP = "http://www.w3.org/2004/08/xop/include"
    zeep.ns.XMIME5 = "http://www.w3.org/2005/05/xmlmime"
    # attach method for the client.
    zeep.Client.attach = _client_attach

def set_attachnode(node):
    """Set the attachment node"""
    cid = node.text[len(FILETAG):]
    node.text = None
    etree.SubElement(
        node, '{{{}}}Include'.format(zeep.ns.XOP), href="cid:{}".format(cid))
    return cid

def get_multipart():
    """Get the main multipart object"""
    part = MIMEMultipart(
        'related', charset='utf-8', type='application/xop+xml',
        boundary=BOUND, start='<soap-env:Envelope>')
    part.set_param('start-info', 'text/xml')

    part.add_header('Accept-Encoding', 'gzip,deflate')
    part.add_header('SOAPAction', '""')
    part.add_header('MIME-Version', '1.0')
    part.add_header('Host', 'test-idahe.ordre.medecin.fr')
    part.add_header('Connection', 'Keep-Alive')
    part.add_header('User-Agent', 'Apache-HttpClient/4.1.1 (java 1.5)')
    return part

def get_envelopepart(envelope):
    """The Envelope part"""
    part = MIMEApplication(etree_to_string(envelope), 'xop+xml', encode_7or8bit)
    part.set_param('charset', 'UTF-8')
    part.set_param('type', 'application/soap+xml')
    part.add_header('Content-Transfer-Encoding', '8bit')
    part.add_header('Content-ID', '<soap-env:Envelope>')
    return part

def _client_attach(self, filename):
    """add attachment"""
    attach = Attachment(filename)
    self.transport.attachments[attach.cid] = attach
    return attach.tag


class Attachment(object):
    """Attachment class"""
    def __init__(self, filename):
        self.filename = filename
        self.basename = basename(self.filename)
        self.cid = func_getid(ID_LEN) + ":" + self.basename
        self.tag = FILETAG + self.cid


class TransportWithAttach(Transport):
    """Transport with attachment"""

    def __init__(self, cache=None, timeout=300, operation_timeout=None,
                 session=None):
        self.attachments = {}
        patch_things()
        super(TransportWithAttach, self).__init__(
            cache=cache, timeout=timeout, operation_timeout=operation_timeout,
            session=session)

    def post_xml(self, address, envelope, headers):
        # Search for values that startswith FILETAG
        filetags = envelope.xpath(
            "//*[starts-with(text(), '{}')]".format(FILETAG))
        # if there is some attached file we set the attachments
        if filetags:
            message = self.set_attachs(filetags, envelope, headers)
        # else just the envelope
        else:
            message = etree_to_string(envelope)
        # post the data.
        return self.post(address, message, headers)

    def set_attachs(self, filetags, envelope, headers):
        """Set mtom attachs and return the right envelope"""
        # let's get the mtom multi part.
        mtom_part = get_multipart()
        # let's set xop:Include for al the files.
        # we need to do this before get the envelope part.
        files = [set_attachnode(f) for f in filetags]
        # get the envelope part.
        env_part = get_envelopepart(envelope)
        # attach the env_part to the multipart.
        mtom_part.attach(env_part)
        # for each filename in files.
        for cid in files:
            # attach the filepart to the multipart.
            part = self.get_attachpart(cid)
            mtom_part.attach(part)
        # some other stuff.
        bound = '--{}'.format(mtom_part.get_boundary())
        marray = mtom_part.as_string().split(bound)
        mtombody = bound
        mtombody += bound.join(marray[1:])
        mtom_part.add_header("Content-Length", str(len(mtombody)))
        headers.update(dict(mtom_part.items()))
        message = mtom_part.as_string().split('\n\n', 1)[1]
        message = message.replace('\n', '\r\n', 5)
        # return the messag for the post.
        return message

    def get_attachpart(self, cid):
        """The file part"""
        attach = self.attachments[cid]
        mime_type = mimetypes.types_map[".%s" % attach.basename.split(".")[-1]].split("/")
        #part = MIMEBase("*", "*")
        part = MIMEBase(mime_type[0], mime_type[1])
        part['Content-Type'] = "%s/%s; charset=us-ascii; name=%s" % (mime_type[0], mime_type[1], attach.filename.split("/")[-1])
        part['Content-Transfer-Encoding'] = "7bit"
        part['Content-ID'] = "<{}>".format(attach.cid)
        part['Content-Disposition'] = 'attachment; name="%s"; filename="%s"' % (attach.filename.split("/")[-1], attach.filename.split("/")[-1])

        file_content = codecs.open(attach.filename, 'r', 'latin-1').read()

        self.file_content = file_content

        part.set_payload(file_content)
        del part['mime-version']
        return part



# -*- coding: utf-8 -*-
"""
======================================================
Test_mtom_attachment :mod:`tests.test_mtom_attachment`
======================================================

"""
import os
from os.path import join, basename
import tempfile
import shutil
import email
import hashlib
from lxml import etree
import pytest
import requests_mock
from zeep import client
from zeep import ns
from zeep import transport_with_attach as twa
TMP_DIR = tempfile.gettempdir()
TMP_OUT = join(TMP_DIR, 'mtom_test', 'out')
TMP_IN = join(TMP_DIR, 'mtom_test', 'in')
MB = 2**20
RESPONSE = b"""
<?xml version="1.0"?>
<soapenv:Envelope
    xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
    xmlns:exam="http://example.com/">
   <soapenv:Header/>
   <soapenv:Body>
      <exam:uploadResponse>
         <message>ok</message>
      </exam:uploadResponse>
   </soapenv:Body>
</soapenv:Envelope>""".strip()

CLIENT = client.Client(
    'tests/wsdl_files/mtom_attachment.wsdl', transport=twa.TransportWithAttach())


@pytest.mark.requests
def create_service():
    """Create service"""
    return CLIENT.create_service(
        '{http://test.ellethee.org/}MtomAttachmentBinding',
        'http://test.python-zeep.org/x')


def create_random_file(filename, size=1024):
    """Create random file"""
    with open(filename, 'wb') as fout:
        fout.write(os.urandom(size))


def get_parts(string):
    """get parts"""
    msg = twa.get_multipart()
    msg.set_payload(string)
    msg = email.message_from_string(msg.as_string())
    parts = []
    for part in msg.walk():
        if part.get_content_maintype() == 'multipart':
            continue
        cid = part.get("Content-ID").strip("<>")
        if cid:
            item = {
                'cid': cid,
                'cte': part.get("Content-Transfer-Encoding"),
            }
            if item['cte'] == 'binary':
                item['payload'] = join(TMP_IN, cid[twa.ID_LEN + 1:])
                with open(item['payload'], 'wb') as fobj:
                    fobj.write(part.get_payload())
            else:
                item['payload'] = part.get_payload()
            parts.append(item)
    return parts


def create_tmp_dirs():
    """create_tmp_dirs"""
    try:
        os.makedirs(TMP_OUT)
    except OSError:
        pass
    try:
        os.makedirs(TMP_IN)
    except OSError:
        pass


def remove_tmp_dirs():
    """removes tmp dirs"""
    shutil.rmtree(TMP_OUT)
    shutil.rmtree(TMP_IN)


def get_file_md5(filename, blocksize=2**20):
    """get file md5"""
    md5 = hashlib.md5()
    with open(filename, "rb") as fin:
        while True:
            buf = fin.read(blocksize)
            if not buf:
                break
            md5.update(buf)
    return md5.hexdigest()


def test_multi_upload():
    """Test multiUpload"""
    service = create_service()
    create_tmp_dirs()
    files = [1, 2, 3, 10]
    for idx, size in enumerate(files):
        filename = join(TMP_OUT, 'test_file_{}'.format(idx))
        create_random_file(filename, size * MB)
        files[idx] = {'fileName': basename(filename),
                      "fileBytes": CLIENT.attach(filename)}
    with requests_mock.mock() as rmock:
        rmock.post('http://test.python-zeep.org/x', text=RESPONSE)
        result = service.multiUpload(files)
        assert result == "ok"
        parts = get_parts(rmock.request_history[0].body)
        xml = etree.fromstring(parts[0]['payload'])
        items = xml.findall(".//{http://test.ellethee.org/}arrayOfUpload/item")
        for item in items:
            filename = item.find("fileName").text
            assert get_file_md5(
                join(TMP_IN, filename)) == get_file_md5(join(TMP_OUT, filename))
        remove_tmp_dirs()


def test_upload():
    """Test Upload"""
    service = create_service()
    create_tmp_dirs()
    filename = join(TMP_OUT, 'test_file')
    create_random_file(filename, 15 * MB)
    with requests_mock.mock() as rmock:
        rmock.post('http://test.python-zeep.org/x', text=RESPONSE)
        result = service.upload(basename(filename), CLIENT.attach(filename))
        assert result == "ok"
        parts = get_parts(rmock.request_history[0].body)
        xml = etree.fromstring(parts[0]['payload'])
        items = xml.findall(".//{http://test.ellethee.org/}arrayOfUpload/item")
        for item in items:
            filename = item.find("fileName").text
            assert get_file_md5(
                join(TMP_IN, filename)) == get_file_md5(join(TMP_OUT, filename))
        remove_tmp_dirs()



from six.moves import range

from zeep import ns


class NamePrefixGenerator(object):
    def __init__(self, prefix='_value_'):
        self._num = 1
        self._prefix = prefix

    def get_name(self):
        retval = '%s%d' % (self._prefix, self._num)
        self._num += 1
        return retval


class UniqueNameGenerator(object):
    def __init__(self):
        self._unique_count = {}

    def create_name(self, name):
        if name in self._unique_count:
            self._unique_count[name] += 1
            return '%s__%d' % (name, self._unique_count[name])
        else:
            self._unique_count[name] = 0
            return name


def max_occurs_iter(max_occurs, items=None):
    assert max_occurs is not None
    generator = range(0, max_occurs if max_occurs != 'unbounded' else 2**31-1)

    if items is not None:
        for i, sub_kwargs in zip(generator, items):
            yield sub_kwargs
    else:
        for i in generator:
            yield i


def create_prefixed_name(qname, schema):
    if not qname:
        return

    if schema and qname.namespace:
        prefix = schema.get_shorthand_for_ns(qname.namespace)
        if prefix:
            return '%s:%s' % (prefix, qname.localname)
    elif qname.namespace in ns.NAMESPACE_TO_PREFIX:
        prefix = ns.NAMESPACE_TO_PREFIX[qname.namespace]
        return '%s:%s' % (prefix, qname.localname)

    if qname.namespace:
        return qname.text
    return qname.localname
